import pygame
import pygame_ecs
import random


class Position(pygame_ecs.BaseComponent):
    def __init__(self, x: int, y: int):
        super().__init__()
        self.x = x
        self.y = y


class BallRenderer(pygame_ecs.BaseComponent):
    def __init__(self, radius: int, color) -> None:
        super().__init__()
        self.radius = radius
        self.color = color


class BallDrawSystem(pygame_ecs.BaseSystem):
    def __init__(self, screen) -> None:
        super().__init__(required_component_types=[Position, BallRenderer])
        self.screen = screen

    def update(self, entity_components):
        pos: Position = entity_components[Position]
        ball_renderer: BallRenderer = entity_components[BallRenderer]
        pygame.draw.circle(
            self.screen, ball_renderer.color, (pos.x, pos.y), ball_renderer.radius
        )


screen = pygame.display.set_mode((800, 600))
clock = pygame.time.Clock()

entity_manager = pygame_ecs.EntityManager()
component_manager = pygame_ecs.ComponentManager()
system_manager = pygame_ecs.SystemManager()
ball_draw_system = BallDrawSystem(screen)
component_manager.init_components()

entities = []
for _ in range(100):
    center = (
        random.randint(0, screen.get_width()),
        random.randint(0, screen.get_height()),
    )
    radius = random.randint(4, 18)
    color = [random.randint(0, 255) for _ in range(3)]
    vel = pygame.math.Vector2(
        (random.random() - 0.5) * 400 / 1000,
        (random.random() - 0.5) * 400 / 1000,
    )
    entity = entity_manager.add_entity()
    component_manager.add_component(entity, Position(center[0], center[1]))
    component_manager.add_component(entity, BallRenderer(radius, color))
    entities.append(entity)

while True:
    for event in pygame.event.get():
        if event.type == pygame.QUIT:
            raise SystemExit

    system_manager.update_entities(entities, component_manager, ball_draw_system)
    pygame.display.update()
    clock.tick(60)
    pygame.display.set_caption(f"FPS: {clock.get_fps()}")



class EntityDoesNotHaveComponent(Exception):
    def __init__(self, entity, component_type) -> None:
        msg = f"Entity of id {entity} doesn't have component of type {component_type} that can be accessed."
        super().__init__(msg)



from timeit import timeit
import pygame_ecs
import random
from sys import argv

try:
    cmds = argv[1]
    if cmds != "perfect":
        cmds = "imperfect"
except IndexError:
    cmds = "perfect"

WIDTH = 800
HEIGHT = 800
ENTITY_AMOUNT = 1_000 * 5


class Position(pygame_ecs.BaseComponent):
    def __init__(self, x: int, y: int):
        super().__init__()
        self.x = x
        self.y = y


class Velocity(pygame_ecs.BaseComponent):
    def __init__(self, vec: list[int | float]) -> None:
        super().__init__()
        self.vec = vec


class BallPhysics(pygame_ecs.BaseSystem):
    def __init__(self) -> None:
        super().__init__(required_component_types=[Position, Velocity])
        self.dt = 0

    def update(self, entity_components):
        pos: Position = entity_components[Position]  # type: ignore
        vel: Velocity = entity_components[Velocity]  # type: ignore
        pos.x += vel.vec[0]  # type: ignore
        pos.y += vel.vec[1]  # type: ignore
        if pos.x > WIDTH or pos.x < 0:
            vel.vec[0] *= -1
        if pos.y > HEIGHT or pos.y < 0:
            vel.vec[1] *= -1


entities = []
entity_manager = pygame_ecs.EntityManager()
component_manager = pygame_ecs.ComponentManager()
system_manager = pygame_ecs.SystemManager()
ball_physics = BallPhysics()
component_manager.init_components()

for _ in range(ENTITY_AMOUNT):
    center = (
        random.randint(0, WIDTH),
        random.randint(0, HEIGHT),
    )
    radius = random.randint(2, 12)
    color = [random.randint(0, 255) for _ in range(3)]
    vel = [
        (random.random() - 0.5) * 400 / 1000,
        (random.random() - 0.5) * 400 / 1000,
    ]
    entity = entity_manager.add_entity()
    component_manager.add_component(entity, Position(center[0], center[1]))
    if cmds[0] == "imperfect":
        component_manager.add_component(entity, Velocity(vel))
    entities.append(entity)

for _ in range(1_000):
    ent = entities[random.randint(0, len(entities) - 1)]
    entity_manager.kill_entity(component_manager, ent)
    entities.remove(ent)

for _ in range(
    2_000
):  # ensure that killing and then spawning entities doesnt break anything
    center = (
        random.randint(0, WIDTH),
        random.randint(0, HEIGHT),
    )
    radius = random.randint(2, 12)
    color = [random.randint(0, 255) for _ in range(3)]
    vel = [
        (random.random() - 0.5) * 400 / 1000,
        (random.random() - 0.5) * 400 / 1000,
    ]
    entity = entity_manager.add_entity()
    component_manager.add_component(entity, Position(center[0], center[1]))
    if cmds[0] == "imperfect":
        component_manager.add_component(entity, Velocity(vel))
    entities.append(entity)

REPEAT = 1_000

res = timeit(lambda: system_manager.update_entities(entities, component_manager, ball_physics), number=REPEAT)  # type: ignore
print(
    f"Took {res/REPEAT} roughly for each frame, using {len(entities)} entities, setting: {cmds}"
)



import typing
from pygame_ecs.components.base_component import BaseComponent
from pygame_ecs.managers.component_manager import ComponentInstanceType


class BaseSystem:
    # typing.Type specifies that it will take subclasses of this class
    def __init__(
        self, required_component_types: list[typing.Type[BaseComponent]]
    ) -> None:
        self.required_component_types = required_component_types

    def update(
        self, entity_components: dict[typing.Type[BaseComponent], ComponentInstanceType]
    ):
        pass

    def __str__(self) -> str:
        return f"<{type(self).__name__}>"



import typing

from pygame_ecs.systems.base_system import BaseSystem
from pygame_ecs.managers.component_manager import ComponentManager

SystemType = typing.TypeVar("SystemType", bound=BaseSystem)


class SystemManager:
    def __init__(self) -> None:
        self.systems: list[BaseSystem] = []

    def add_system(self, system: BaseSystem):
        self.systems.append(system)

    def remove_system(self, system: BaseSystem):
        self.systems.remove(system)

    def update_entities(
        self, entities, component_manager: ComponentManager, system: SystemType
    ):
        for entity in entities:
            has_components = True
            components_to_give = {}
            for comp_type in system.required_component_types:
                try:
                    comp = component_manager.components[comp_type][entity]
                    components_to_give[type(comp)] = comp
                except KeyError:
                    has_components = False
                    break
            if has_components:
                system.update(components_to_give)



class Entity(object):
    __slots__ = ("_id",)

    def __init__(self, id):
        self._id = id

    def __repr__(self):
        return f"{type(self).__name__} {self._id}"

    def __hash__(self):
        return self._id

    def __eq__(self, other):
        return self._id == hash(other)



import pygame
import random
import pygame_ecs
from pygame._sdl2 import Renderer, Texture, Window, get_drivers

pygame.init()

ENTITY_AMOUNT = 1_000 * 10
WIDTH = 800
HEIGHT = 600


class Position(pygame_ecs.BaseComponent):
    __slots__ = ("x", "y")

    def __init__(self, x: int, y: int):
        super().__init__()
        self.x = x
        self.y = y


class BallRenderer(pygame_ecs.BaseComponent):
    __slots__ = ("radius", "color")

    def __init__(self, radius: int, color) -> None:
        super().__init__()
        self.radius = radius
        self.color = color


class Velocity(pygame_ecs.BaseComponent):
    __slots__ = "vec"

    def __init__(self, vec: pygame.math.Vector2) -> None:
        super().__init__()
        self.vec = vec


class BallDrawSystem(pygame_ecs.BaseSystem):
    def __init__(self, texture: Texture) -> None:
        super().__init__(required_component_types=[Position, BallRenderer])
        self.texture = texture

    def update(self, entity_components):
        pos: Position = entity_components[Position]  # type: ignore
        ball_renderer: BallRenderer = entity_components[BallRenderer]  # type: ignore

        self.texture.color = ball_renderer.color  # type: ignore
        self.texture.draw(
            None, (pos.x, pos.y, ball_renderer.radius, ball_renderer.radius)
        )


class BallPhysics(pygame_ecs.BaseSystem):
    def __init__(self, screen) -> None:
        super().__init__(required_component_types=[Position, Velocity])
        self.dt = 0
        self.screen = screen

    def update(self, entity_components):
        pos: Position = entity_components[Position]  # type: ignore
        vel: Velocity = entity_components[Velocity]  # type: ignore
        pos.x += vel.vec.x * self.dt  # type: ignore
        pos.y += vel.vec.y * self.dt  # type: ignore
        if pos.x > WIDTH or pos.x < 0:
            vel.vec.x *= -1
        if pos.y > HEIGHT or pos.y < 0:
            vel.vec.y *= -1


screen = pygame.display.set_mode((WIDTH, HEIGHT))
window = Window.from_display_module()


print(list(get_drivers()))
# change based on drivers
renderer = Renderer(window, index=3, accelerated=1)


texture = Texture.from_surface(renderer, pygame.image.load("test/circle.png"))

entities = []
entity_manager = pygame_ecs.EntityManager()
component_manager = pygame_ecs.ComponentManager()
system_manager = pygame_ecs.SystemManager()
ball_physics = BallPhysics(screen)
ball_draw_system = BallDrawSystem(texture=texture)

component_manager.init_components()


for _ in range(ENTITY_AMOUNT):
    center = (
        random.randint(0, WIDTH),
        random.randint(0, HEIGHT),
    )
    radius = random.randint(5, 15)
    color = [random.randint(0, 255) for _ in range(3)]
    color.append(255)
    vel = pygame.math.Vector2(
        (random.random() - 0.5) * 400 / 1000,
        (random.random() - 0.5) * 400 / 1000,
    )
    entity = entity_manager.add_entity()
    component_manager.add_component(entity, Position(center[0], center[1]))
    if random.randint(0, 1):
        component_manager.add_component(entity, Velocity(vel))
    component_manager.add_component(entity, BallRenderer(radius, color))
    entities.append(entity)


clock = pygame.time.Clock()
dt = 0

renderer.draw_color = (
    0,
    0,
    0,
    255,
)  # type: ignore
# renderer.draw_color is used for clearing the screen and drawing primitives

running = True
while running:
    for event in pygame.event.get():
        if event.type == pygame.QUIT:
            running = False

        if event.type == pygame.KEYDOWN and event.key == pygame.K_ESCAPE:
            running = False
    pygame.display.set_caption(f"FPS: {clock.get_fps()}")
    ball_physics.dt = dt
    system_manager.update_entities(entities, component_manager, ball_physics)

    renderer.clear()
    system_manager.update_entities(entities, component_manager, ball_draw_system)
    renderer.present()
    dt = clock.tick(60)

pygame.quit()



from pygame_ecs.entity import Entity
from pygame_ecs.managers.component_manager import ComponentManager


class EntityManager:
    def __init__(self) -> None:
        self.count = 0
        self.dead_entities: list[Entity] = []

    def add_entity(self):
        if len(self.dead_entities) > 0:
            entity = self.dead_entities.pop()
        else:
            entity = Entity(self.count)
            self.count += 1
        return entity

    def kill_entity(self, component_manager: ComponentManager, entity: Entity):
        self.dead_entities.append(entity)
        for component_type in component_manager.components.keys():
            try:
                del component_manager.components[component_type][entity]
            except KeyError:
                pass



class BaseComponent:
    """An Base class that all Components must inherit from."""

    def __init__(self) -> None:
        pass

    def __str__(self) -> str:
        return f"{type(self).__name__} {self.__dict__()}"



import typing

from pygame_ecs.components.base_component import BaseComponent
from pygame_ecs.entity import Entity
from pygame_ecs.exceptions import EntityDoesNotHaveComponent


ComponentInstanceType = typing.TypeVar("ComponentInstanceType", bound=BaseComponent)


class ComponentManager:
    def __init__(self) -> None:
        self.components: dict[typing.Type[BaseComponent], dict[Entity, ComponentInstanceType]] = {}  # type: ignore

    def init_components(self):
        # get all subclasses using BaseComponent
        component_subclasses = BaseComponent.__subclasses__()
        for component_subclass in component_subclasses:
            self.components[component_subclass] = {}

    def add_component(self, entity, component):
        self.components[type(component)][entity] = component

    def remove_component(self, entity, component_type):
        try:
            del self.components[component_type][entity]
        except KeyError:
            raise EntityDoesNotHaveComponent(entity, component_type)



from pygame_ecs.components.base_component import BaseComponent
from pygame_ecs.entity import Entity
from pygame_ecs.managers.component_manager import ComponentManager
from pygame_ecs.managers.entity_manager import EntityManager
from pygame_ecs.managers.system_manager import SystemManager
from pygame_ecs.systems.base_system import BaseSystem




import random
import pygame
import pygame_ecs


class Position(pygame_ecs.BaseComponent):
    def __init__(self, x: int, y: int):
        super().__init__()
        self.x = x
        self.y = y


class BallRenderer(pygame_ecs.BaseComponent):
    def __init__(self, radius: int, color) -> None:
        super().__init__()
        self.radius = radius
        self.color = color


class Velocity(pygame_ecs.BaseComponent):
    def __init__(self, vec: pygame.math.Vector2) -> None:
        super().__init__()
        self.vec = vec


class BallDrawSystem(pygame_ecs.BaseSystem):
    def __init__(self, screen) -> None:
        super().__init__(required_component_types=[Position, BallRenderer])
        self.screen = screen

    def update(self, entity_components):
        pos: Position = entity_components[Position]  # type: ignore
        ball_renderer: BallRenderer = entity_components[BallRenderer]  # type: ignore
        pygame.draw.circle(self.screen, ball_renderer.color, (pos.x, pos.y), ball_renderer.radius)  # type: ignore


class BallPhysics(pygame_ecs.BaseSystem):
    def __init__(self, screen) -> None:
        super().__init__(required_component_types=[Position, Velocity])
        self.dt = 0
        self.screen = screen

    def update(self, entity_components):
        pos: Position = entity_components[Position]  # type: ignore
        vel: Velocity = entity_components[Velocity]  # type: ignore
        pos.x += vel.vec.x * self.dt  # type: ignore
        pos.y += vel.vec.y * self.dt  # type: ignore
        if pos.x > self.screen.get_width() or pos.x < 0:
            vel.vec.x *= -1
        if pos.y > self.screen.get_height() or pos.y < 0:
            vel.vec.y *= -1


class App:
    def __init__(self) -> None:
        self.screen = pygame.display.set_mode((800, 600))
        self.clock = pygame.time.Clock()
        self.dt = 0

    def setup(self):
        self.entities = []
        self.entity_manager = pygame_ecs.EntityManager()
        self.component_manager = pygame_ecs.ComponentManager()
        self.system_manager = pygame_ecs.SystemManager()
        self.ball_draw_system = BallDrawSystem(self.screen)
        self.ball_physics = BallPhysics(self.screen)
        self.component_manager.init_components()

    def add_entities(self):
        for _ in range(1_000 * 10):
            center = (
                random.randint(0, self.screen.get_width()),
                random.randint(0, self.screen.get_height()),
            )
            radius = random.randint(2, 12)
            color = [random.randint(0, 255) for _ in range(3)]
            vel = pygame.math.Vector2(
                (random.random() - 0.5) * 400 / 1000,
                (random.random() - 0.5) * 400 / 1000,
            )
            entity = self.entity_manager.add_entity()
            self.component_manager.add_component(entity, Position(center[0], center[1]))
            self.component_manager.add_component(entity, BallRenderer(radius, color))
            if random.randint(0, 1):
                self.component_manager.add_component(entity, Velocity(vel))
            self.entities.append(entity)

    def draw(self):
        self.screen.fill("black")
        self.system_manager.update_entities(
            self.entities, self.component_manager, self.ball_draw_system
        )

    def run(self):
        self.setup()
        self.add_entities()
        while True:
            pygame.display.set_caption(f"FPS: {self.clock.get_fps()}")
            self.get_input()
            self.update()
            self.draw()
            self.dt = self.clock.tick(60)
            pygame.display.update()

    def update(self):
        self.ball_physics.dt = self.dt
        self.system_manager.update_entities(
            self.entities, self.component_manager, self.ball_physics
        )

    def get_input(self):
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                raise SystemExit


if __name__ == "__main__":
    app = App()
    app.run()
